unit LightMap;
{ Dynamic Light demo (c)2002 by Paul TOTH <tothpaul@free.fr>
  http://tothpaul.free.fr

  based on the C demo

  =============
  Fake Dynamic Lights

  Author : Nate Miller
  Date   : 07/01/01
  Email  : nathanm@uci.edu
  Web    : http://nate.scuzzy.net
  =============
}


{
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
}
interface

uses
 Classes,SysUtils,Graphics,DelphiGL;

Type
 TVector2D=packed record
  u,v:single;
 end;

 TVector3D=packed record
  x,y,z:single;
 end;

 TPlane=packed record
  pad:integer; // probably the virtual destructor ... see C source
  n:TVector3D;
  d:single;
 end;

 TGeomHeader=packed record
  VertexCount  :integer;
  FaceCount    :integer;
  TextureCount :integer;
  LightmapCount:integer;
 // Vertices[VertexCount] of TVertex;
 // Faces[FaceCount] of TFace;
 // Textures[TextureCount] of TTexture;
 // Lightmaps[LightmapCount] of TLightmap;
 end;

 TVertex=packed record
  tv :TVector2D; // texture coordinates
  lv :TVector2D; // lightmap coordinates
  pos:TVector3D; // vertex position
 end;

 TFace=packed record
  start:integer; // first face vert in vertex array
  count:integer; // number of verts in the face
  id   :integer; // texture index into the texture array
  lid  :integer; // lightmap index info the lightmap array
  p    :TPlane;
 end;

 TTexture=packed record
  id  :integer;
  name:array[0..63] of char;
 end;

 TLightmap=packed record
  id :integer;
  map:packed array[0..127,0..127,0..2] of byte; // 128x128 rgb
 end;

 TMap=class
  Vertices :array of TVertex;
  Faces    :array of TFace;
  Textures :array of TTexture;
  Lightmaps:array of TLightmap;
  Constructor Create(AFileName:string);
  Procedure LoadTextures;
  procedure RenderLightmap;
  procedure RenderTextures;
  procedure RenderWire;
  procedure RenderNormals;
 end;

const
 dLightBase=1;

Type
 TDLightBase=class
  lightType:integer;
  size:single;
  pos:TVector3D;
  color:TVector3D;
  Constructor Create;
  procedure Frame(time:single); virtual; abstract;
 end;

 TDLightPulsar=class(TDLightBase)
  min,max,amp:single;
  Constructor Create(asize:single; x,y,z:single; r,g,b:single; amin,amax:single; aamp:single);
  procedure Frame(time:single); override;
 end;

 TDLightGen=class
  up,right:array of TVector3D;
  texId:integer;
  lights:array of TDLightBase;
  constructor Create(map:TMap);
  destructor destroy; override;
  procedure Frame(map:TMap);
  procedure add(light:TDLightBase);
 end;

var
 dlight:TDLightGen;


var
 map:TMap;

procedure adjustcolors(p:pointer; w,h,bpp:integer);

implementation

procedure adjustcolors(p:pointer; w,h,bpp:integer);
 var
  b:^byte;
  x,y,z:integer;
 begin
  b:=p;
  for y:=1 to h do begin
   for x:=1 to w do begin
    for z:=1 to bpp do begin
     if (b^)>128 then b^:=255 else b^:=2*b^;
     inc(b);
    end;
   end;
  end;
 end;

Const
 axis:array[0..2] of TVector3D=(
  (x:1.0;y:0.0;z:0.0),
  (x:0.0;y:1.0;z:0.0),
  (x:0.0;y:0.0;z:1.0)
 );

procedure Normalize(var n:TVector3D);
 var
  d:double;
  l:single;
 begin
  d:=(n.x*n.x)+(n.y*n.y)+(n.z*n.z);
  l:=sqrt(d);
  if l<>0 then begin
   l:=1/l;
   n.x:=n.x*l;
   n.y:=n.y*l;
   n.z:=n.z*l;
  end;
 end;

procedure u3dGetRightUp(const n:TVector3D; var right,up:TVector3D);
 var
  fn:array[0..2] of single;
  major:integer;
 begin
  fn[0]:=abs(n.x);
  fn[1]:=abs(n.y);
  fn[2]:=abs(n.z);
  major:=0;
 // find the major axis
  if (fn[1] > fn[major]) then  major := 1;
  if (fn[2] > fn[major]) then  major := 2;
 // build right vector by hand
  if (fn[0] = 1) or (fn[1] = 1) or (fn[2] = 1) then begin
   if (major = 0) and (n.x > 0) then begin
    right.x:=0;
    right.y:=0;
    right.z:=-1;
   end else if (major = 0) then begin
    right.x:=0;
    right.y:=0;
    right.z:=1;
   end;
   if (major = 1) or ((major = 2) and (n.z > 0)) then begin
    right.x:=1;
    right.y:=0;
    right.z:=0;
   end;

   if (major = 2) and (n.z < 0.0) then begin
    right.x:=-1;
    right.y:=0;
    right.z:=0;
   end;
  end else begin
   right.x:=(axis[major].y*n.z)-(axis[major].z*n.y);
   right.y:=(axis[major].z*n.x)-(axis[major].x*n.z);
   right.z:=(axis[major].x*n.y)-(axis[major].y*n.x);
  end;

  up.x:=(n.y*right.z)-(n.z*right.y);
  up.y:=(n.z*right.x)-(n.x*right.z);
  up.z:=(n.x*right.y)-(n.y*right.x);

  Normalize(right);
  Normalize(up);
 end;

function dot(const v1,v2:TVector3D):single;
 begin
  result:=v1.x*v2.x+v1.y*v2.y+v1.z*v2.z;
 end;

function Distance(const p:TPlane;const v:TVector3D):single;
 begin
  result:=dot(p.n,v)+p.d;
 end;

procedure u3dProjectVector(const src,v:TVector3D; var res:TVector3D);
 var
  scale:single;
 begin
  scale:=dot(v,src)/dot(v,v);
  res.x:=v.x*scale;
  res.y:=v.y*scale;
  res.z:=v.z*scale;
 end;

procedure u3dProjectPlane(const src,v1,v2:TVector3D; var res:TVector3D);
 var
  t1,t2:TVector3D;
 begin
  u3dProjectVector(src,v1,t1);
  u3dProjectVector(src,v2,t2);
  res.x:=t1.x+t2.x;
  res.y:=t1.y+t2.y;
  res.z:=t1.z+t2.z;
 end;

Constructor TMap.Create(AFileName:string);
 var
  s:TFileStream;
  geom:TGeomHeader;
  i:integer;
  b:TBitmap;
 begin
  s:=TFileStream.Create(AFileName,fmOpenRead or fmShareDenyNone);
  s.ReadBuffer(geom,SizeOf(geom));
  SetLength(Vertices,geom.VertexCount);   s.ReadBuffer(Vertices[0],geom.VertexCount*SizeOf(TVertex));
  setLength(Faces,geom.FaceCount);        s.ReadBuffer(Faces[0],geom.FaceCount*SizeOf(TFace));
  setLength(Textures,geom.TextureCount);  s.ReadBuffer(Textures[0],geom.TextureCount*SizeOf(TTexture));
  SetLength(Lightmaps,geom.LightmapCount);s.ReadBuffer(Lightmaps[0],geom.LightmapCount*SizeOf(TLightmap));
  s.Free;

  b:=TBitmap.Create;
  b.PixelFormat:=pf24bit;
  b.Width:=128;
  b.Height:=128;
  for i:=0 to length(Lightmaps)-1 do begin
   move(Lightmaps[i].map,b.ScanLine[127]^,128*128*3);
   b.SaveToFile('lightmap'+IntToStr(i)+'.bmp');
  end;
  b.Free;

  LoadTextures;

  for i:=0 to length(Lightmaps)-1 do begin
    AdjustColors(@lightmaps[i].map, 128, 128, 3);
    glGenTextures(1, @lightmaps[i].id);
    glBindTexture(GL_TEXTURE_2D, lightmaps[i].id);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glTexImage2D(GL_TEXTURE_2D, 0, 3, 128, 128, 0, GL_RGB, GL_UNSIGNED_BYTE, @lightmaps[i].map);
  end;
 end;

Procedure TMap.LoadTextures;
 var
  b:TBitmap;
  s:string;
  i:integer;
  components:integer;
  format:integer;
 begin
  b:=TBitmap.Create;
  try
   for i:=0 to Length(Textures)-1 do begin
    s:=Textures[i].Name+'.bmp';
    if FileExists(s) then begin
     b.LoadFromFile(s);
     case b.PixelFormat of
      pf8Bit :components:=1;
      pf24bit:components:=3;
      pf32bit:components:=4;
     end;
     AdjustColors(b.ScanLine[b.Height-1],b.width,b.height,components);
     glGenTextures(1,@Textures[i].id);
     glBindTexture(GL_TEXTURE_2D, Textures[i].id);
     glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
     glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
     glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
     glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
     glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
     glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
     case components of
      1: format:=GL_LUMINANCE;
      3: format:=GL_RGB;
      4: format:=GL_RGBA;
     end;
     gluBuild2DMipmaps(GL_TEXTURE_2D, components, b.width, b.height, format, GL_UNSIGNED_BYTE, b.ScanLine[b.Height-1]);
    end;
   end;
  finally
   b.Free;
  end;
 end;

procedure TMap.RenderLightmap;
 var
  i:integer;
 begin
  glColor4f(1.0, 1.0, 1.0, 1.0);
  glEnableClientState(GL_VERTEX_ARRAY);
  glEnableClientState(GL_TEXTURE_COORD_ARRAY);

  glVertexPointer(3, GL_FLOAT, sizeof(TVertex), vertices[0].pos);
  glTexCoordPointer(2, GL_FLOAT, sizeof(TVertex), vertices[0].lv);

  // Draw all of our faces
  for i:=0 to length(Faces)-1 do begin
   glBindTexture(GL_TEXTURE_2D, lightmaps[faces[i].lid].id);
   glDrawArrays(GL_TRIANGLE_FAN , faces[i].start ,faces[i].count);
  end;

  glDisableClientState(GL_TEXTURE_COORD_ARRAY);
  glDisableClientState(GL_VERTEX_ARRAY);
 end;

procedure TMap.RenderTextures;
 var
  i:integer;
 begin
  glColor4f(1.0, 1.0, 1.0, 1.0);
  glEnableClientState(GL_VERTEX_ARRAY);
  glEnableClientState(GL_TEXTURE_COORD_ARRAY);

  glVertexPointer(3, GL_FLOAT, sizeof(TVertex), Vertices[0].pos);
  glTexCoordPointer(2, GL_FLOAT, sizeof(TVertex), Vertices[0].tv);

  // Draw all of our faces
  for i:=0 to length(Faces)-1 do begin
   glBindTexture(GL_TEXTURE_2D, textures[faces[i].id].id);
   glDrawArrays(GL_TRIANGLE_FAN , faces[i].start ,faces[i].count);
  end;

  glDisableClientState(GL_TEXTURE_COORD_ARRAY);
  glDisableClientState(GL_VERTEX_ARRAY);
 end;

procedure TMap.RenderWire;
 var
  i:integer;
 begin
  glColor4f(1.0, 1.0, 1.0, 1.0);
  glEnableClientState(GL_VERTEX_ARRAY);
  glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
  glDisable(GL_DEPTH_TEST);

  for i:=0 to length(Faces)-1 do
     glDrawArrays(GL_TRIANGLE_FAN , faces[i].start ,faces[i].count);

  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
  glEnable(GL_DEPTH_TEST);
  glDisableClientState(GL_VERTEX_ARRAY);
 end;

procedure TMap.RenderNormals;
var
 i:integer;
begin
   glColor4f(1.0, 1.0, 1.0, 1.0);

   glColor3f(0, 1, 0);
   glBegin(GL_LINES);
  (*
      for i:= to length(Faces)-1 do begin
         vec3_t center;
         for (int k = faces[j].start; k < faces[j].start + faces[j].num; k++)
            center.Add(verts[k].pos);
         center.Scale(1.0f / faces[j].num);

         glVertex3fv(center);
         center += (50 * faces[j].p.n);
         glVertex3fv(center);
      end;
   *)
   glEnd();
end;

Constructor TDLightBase.Create;
 begin
  lightType:=dLightBase;
  size:=0;
 end;

Constructor TDLightGen.Create(Map:TMap);
 var
  n,i:integer;
 begin
  n:=Length(Map.Faces);
  SetLength(up,n);
  SetLength(right,n);
  for i:=0 to n-1 do u3dGetRightUp(Map.Faces[i].p.n,right[i],up[i]);
 end;

Destructor TDLightGen.Destroy;
 var
  i:integer;
 begin
  for i:=0 to length(lights)-1 do lights[i].Free;
 end;

procedure TDLightGen.Frame(Map:TMap);
 var
  num:integer;
  i:integer;
  l:integer;
  d:single;
  n:TVector3D;
  s:single;
  c:single;
  j:integer;
  v:TVector3D;
  uv:TVector2D;
 begin
   num := Length(map.Faces);

   glBindTexture(GL_TEXTURE_2D, texId);

   for i:=0 to num-1 do begin
      for l:=0 to Length(Lights)-1 do begin
       d:=Distance(map.Faces[i].p,Lights[l].Pos);
      // Is the face too far away to be lit by the current light?
       if (d > Lights[l].size) then continue;
      // Project our light position onto the faces plane.  Gives the closest
      // point on the plane to the light position.
       u3dProjectPlane(Lights[l].Pos, right[i], up[i], n);
      // - 'scale' makes the light smaller as it moves away from the face
      // - 'cs' scales the color of the light so it is less intense as it
      //    moves away from the face.
       s:= 1 / ((2 * Lights[l].size) - d);
       c:= 1 - (d / Lights[l].size);
         with Lights[l] do
          glColor4f(color.x * c, color.y * c, color.z * c, 1);
         glBegin(GL_TRIANGLE_FAN);
         for j := map.faces[i].start to map.faces[i].start+map.faces[i].Count-1 do begin
          // Project the light image onto the face
           v.x:=Map.Vertices[j].pos.x-n.x;
           v.y:=Map.Vertices[j].pos.y-n.y;
           v.z:=Map.Vertices[j].pos.z-n.z;
          // Get our texture coordinates
           uv.u := (Dot(v,right[i]) * s) + 0.5;
           uv.v := (Dot(v,up[i]) * s) + 0.5;
           glTexCoord2fv(uv);
           glVertex3fv(map.vertices[j].pos);
         end;
         glEnd();
      end;
   end;
 end;

procedure TDLightGen.Add(light:TDLightBase);
 var
  i:integer;
 begin
  i:=length(lights);
  setlength(lights,i+1);
  lights[i]:=light;
 end;

Constructor TDLightPulsar.Create(asize:single; x,y,z:single; r,g,b:single; amin,amax:single; aamp:single);
 begin
  size:=asize;
  pos.x:=x;
  pos.y:=y;
  pos.z:=z;
  color.x:=r;
  color.y:=g;
  color.z:=b;
  min:=amin;
  max:=amax;
  amp:=aamp;
 end;

procedure TDLightPulsar.Frame(time:single);
 begin
   size :=size + (time * amp);

   if (size > max) then begin
      size := max;
      amp := -amp;
   end;

   if (size < min) then begin
      size := min;
      amp := -amp;
   end;
 end;

end.
