unit dlight1;

{ Dynamic Light demo (c)2002 by Paul TOTH <tothpaul@free.fr>
  http://tothpaul.free.fr

  based on the C demo

  =============
  Fake Dynamic Lights

  Author : Nate Miller
  Date   : 07/01/01
  Email  : nathanm@uci.edu
  Web    : http://nate.scuzzy.net
  =============
}


{
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
}
interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, DelphiGL, ExtCtrls;

const
 path='../';

type
  TOption=(lightmaps,lights,textures,wire,normals);

  TForm1 = class(TForm)
    Timer1: TTimer;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure FormKeyPress(Sender: TObject; var Key: Char);
    procedure FormMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure FormMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure Timer1Timer(Sender: TObject);
  private
    { Dclarations prives }
    fDC:THandle;
    fGL:THandle;
    mm:boolean;
    mx,my:integer;
    rx,ry:single;
    last:integer;
    Options:set of TOption;
    procedure switch(option:TOption);
  public
    { Dclarations publiques }
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}

uses
 Lightmap;

procedure TForm1.Switch(option:TOption);
 begin
  if option in options then exclude(options,option) else include(options,option);
 end;

procedure TForm1.FormCreate(Sender: TObject);
 var
  pfd:TPIXELFORMATDESCRIPTOR;
  pixelformat:integer;
  b:TBitmap;
  components:integer;
  format:integer;
 begin
 // OpenGL code
  fDC:=GetDC(Handle);
  FillChar(pfd,SizeOf(pfd),0);
  pfd.nSize       := sizeof(pfd);
  pfd.nVersion    := 1;
  pfd.dwFlags     := PFD_DRAW_TO_WINDOW or PFD_SUPPORT_OPENGL or PFD_DOUBLEBUFFER;
  pfd.iLayerType  := PFD_MAIN_PLANE;
  pfd.iPixelType  := PFD_TYPE_RGBA;
  pfd.cColorBits  := 32;
  pfd.iLayerType  :=PFD_MAIN_PLANE;
  pixelformat:=ChoosePixelFormat(fDC, @pfd);
  if PixelFormat=0 then RaiseLastWin32Error;
  if not SetPixelFormat(fDC, pixelformat, @pfd) then RaiseLastWin32Error;
  fGL:=wglCreateContext(fDC);
  wglMakeCurrent(fDC,fGL);
 // Application specific code
  map:=TMap.Create('map.geom');
  dlight:=TDLightGen.Create(map);
  dlight.add(TDLightPulsar.Create(150, 0,-64,-192, 1,1,1, 50,250, 50));
  glCullFace(GL_BACK);
  options:=[lightmaps,lights,textures];
  b:=TBitmap.Create;
  try
   b.LoadFromFile('ball.bmp');
   //AdjustColors(b.ScanLine[b.Height-1],b.Width,b.Height,1);
   glGenTextures(1,@dlight.texId);
   glBindTexture(GL_TEXTURE_2D, dlight.texId);
   glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
   case b.pixelFormat of
    pf8bit : begin components:=1; format:=GL_LUMINANCE; end;
    pf24bit: begin components:=3; format:=GL_RGB; end;
    pf32bit: begin components:=4; format:=GL_RGBA; end;
    else exit;
   end;
   gluBuild2DMipmaps(GL_TEXTURE_2D, components, b.width, b.height, format, GL_UNSIGNED_BYTE, b.ScanLine[b.Height-1]);
  finally
   b.Free;
  end;
 end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
 wglMakeCurrent(fDC,0);
 wglDeleteContext(fGL);
 ReleaseDC(Handle,fDC);
end;

procedure TForm1.FormResize(Sender: TObject);
begin
  if (fGL=0)or(ClientHeight=0) then exit;
  glViewport(0, 0, ClientWidth, ClientHeight);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity;
  gluPerspective(64,ClientWidth/ClientHeight,1,4096);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity;
end;

procedure TForm1.FormPaint(Sender: TObject);
begin
 glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);
 glLoadIdentity;

  glTranslatef(0,0,-250);
  glRotatef(rx,1,0,0);
  glRotatef(ry,0,1,0);

  glEnable(GL_TEXTURE_2D);
  glEnable(GL_BLEND);

  glDepthFunc(GL_LESS);
  glDepthMask(GL_TRUE);

  // First pass, lightmaps
  if (lightmaps in options) then begin
   glBlendFunc(GL_ONE, GL_ZERO);
   map.RenderLightmap;
  end;

  glDepthFunc(GL_EQUAL);
  glDepthMask(GL_FALSE);

  // Show dynamic lights?
  if (lights in options) then begin
     // Second pass, dynamic lights
     // They are added to the lightmap values
     glBlendFunc(GL_SRC_ALPHA, GL_ONE);
     dlight.Frame(map);
  end;

  // Third pass, render the world texture
  if (textures in options) then begin
   if (options * [lightmaps,lights]=[]) then begin
    glDisable(GL_BLEND);
    glDisable(GL_DEPTH_TEST);
   end;
   glBlendFunc(GL_DST_COLOR, GL_ZERO);
   map.RenderTextures;
  end;

  glDisable(GL_BLEND);
  glDepthFunc(GL_LESS);
  glDepthMask(GL_TRUE);
  glDisable(GL_TEXTURE_2D);

  // Wireframe?
  if (wire in options) then map.RenderWire;

  // Normals?
  if (normals in options) then map.RenderNormals;

  // Show the dynamic light points?
(*
  if (options['p'])
  {
     dLightList &l = dlight.GetLights();
     dLightListIter itr;

     glBegin(GL_POINTS);
        for (itr = l.begin(); itr != l.end(); ++itr)
        {
           glColor3fv((*itr)->GetColor());
           glVertex3fv((*itr)->GetPos());
        }
     glEnd();
*)
 glFlush;
 SwapBuffers(fDC);
end;

procedure TForm1.FormKeyPress(Sender: TObject; var Key: Char);
begin
 if key='w' then switch(wire);
 if key='n' then switch(normals);
 if key='t' then switch(textures);
 if key='l' then switch(lightmaps);
 if key='d' then switch(lights);
 Paint;
end;

procedure TForm1.FormMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
 mm:=true;
 mx:=x;
 my:=y;
end;

procedure TForm1.FormMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
 if mm then begin
  rx:=rx+(my-y); my:=y;
  ry:=ry+(mx-x); mx:=x;
  paint;
 end;
end;

procedure TForm1.FormMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
 mm:=false;
end;

procedure TForm1.Timer1Timer(Sender: TObject);
var
 t:integer;
 i:integer;
begin
 t:=getTickCount;
 if last<>0 then
  for i:=0 to length(dlight.Lights)-1 do dlight.lights[i].Frame((t-last)/500);
 last:=t;
 paint;
end;

end.
